#include "raylib.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>

#define SIZE_OF_HEX 25

struct celluleDemineur{
    int line, column;
    bool hasMine;
    int nearbyMine;
    Vector2 pos;
    bool isPressed;
    bool ifFlagged;
    bool isNeareast;
    Color couleurCase;
};

typedef struct celluleDemineur * CelluleDemineur;


bool containsMine(int line, int col, CelluleDemineur** matriceMines);

CelluleDemineur* recupererVoisins(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice);

void incrementationVoisinMine(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice);

bool revelerVoisins(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice);

bool revelerCase(CelluleDemineur** matriceMines, int line, int column, int nbLigneMatrice, int nbColonneMatrice);

void generationMines(int nbMines, int nbLigneMatrice, int nbColonneMatrice, CelluleDemineur** matriceMines);

void initGrille(int nbLigneMatrice, int nbColonneMatrice, CelluleDemineur** matriceMines);

float getDistanceBetweenPoints(Vector2 a, Vector2 b);


bool containsMine(int line, int col, CelluleDemineur** matriceMines){
    return (matriceMines[line][col]->hasMine == true);
}

CelluleDemineur* recupererVoisins(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice){
    CelluleDemineur* tabVoisins = calloc(6, sizeof(struct celluleDemineur));
    for(int i=0; i<6; i++){
        tabVoisins[i] = NULL;
    }
    
    bool linePrev = (line-1 >= 0), colPrev = (column-1 >= 0);
    bool lineNext = (line+1 < nbLigneMatrice), colNext = (column+1 < nbColonneMatrice);
    
    if(colPrev){tabVoisins[0] = matriceMines[line][column-1];}
    if(colNext){tabVoisins[1] = matriceMines[line][column+1];}
    
    if(linePrev){tabVoisins[2] = matriceMines[line-1][column];}
    if(lineNext){tabVoisins[3] = matriceMines[line+1][column];}
    
    if(line%2 == 0){
        if(linePrev && colNext){tabVoisins[4] = matriceMines[line-1][column+1];}
        if(lineNext && colNext){tabVoisins[5] = matriceMines[line+1][column+1];}
    }else{
        if(linePrev && colPrev){tabVoisins[4] = matriceMines[line-1][column-1];}
        if(lineNext && colPrev){tabVoisins[5] = matriceMines[line+1][column-1];}
    }
    
    return tabVoisins;
}

void incrementationVoisinMine(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice){
    CelluleDemineur* voisins = recupererVoisins(line, column, matriceMines, nbLigneMatrice, nbColonneMatrice);
    
    for(int i=0; i<6; i++){
        if(voisins[i] != NULL){
            voisins[i]->nearbyMine++;
        }
    }
    free(voisins);
}

bool revelerVoisins(int line, int column, CelluleDemineur** matriceMines, int nbLigneMatrice, int nbColonneMatrice){
    CelluleDemineur* voisins = recupererVoisins(line, column, matriceMines, nbLigneMatrice, nbColonneMatrice);
    bool res = false, resTemp;
    
    for(int i=0; i<6; i++){
        if(voisins[i] != NULL){
            resTemp = revelerCase(matriceMines, voisins[i]->line, voisins[i]->column, nbLigneMatrice, nbColonneMatrice);
            if(resTemp) res = true;
        }
    }
    free(voisins);
    return res;
}

bool revelerCase(CelluleDemineur** matriceMines, int line, int column, int nbLigneMatrice, int nbColonneMatrice){
    bool res = false;
    if(!matriceMines[line][column]->ifFlagged && !matriceMines[line][column]->isPressed){
        matriceMines[line][column]->isPressed = true;
    
        if(matriceMines[line][column]->nearbyMine == 0){
            matriceMines[line][column]->couleurCase = LIGHTGRAY;
            res = revelerVoisins(line, column, matriceMines, nbLigneMatrice, nbColonneMatrice);
        }else if(containsMine(line, column, matriceMines)){
            matriceMines[line][column]->couleurCase = MAROON;
            res = true;
        }else{
            matriceMines[line][column]->couleurCase = LIGHTGRAY;
        }
    }
    return res;
}

void generationMines(int nbMines, int nbLigneMatrice, int nbColonneMatrice, CelluleDemineur** matriceMines){
    int lineRand = rand()%nbLigneMatrice;
    int colRand = rand()%nbColonneMatrice;
    
    for(int i=0; i<nbMines; i++){
        
        while(containsMine(lineRand, colRand, matriceMines)){
            lineRand = rand()%nbLigneMatrice;
            colRand = rand()%nbColonneMatrice;
        }
        
        matriceMines[lineRand][colRand]->hasMine = true;
        matriceMines[lineRand][colRand]->nearbyMine = -9;
        incrementationVoisinMine(lineRand, colRand, matriceMines, nbLigneMatrice, nbColonneMatrice);
        
    }
}

void initGrille(int nbLigneMatrice, int nbColonneMatrice, CelluleDemineur** matriceMines){
    int screenHeight = GetScreenHeight(), screenWidth = GetScreenWidth();
    
    for(int i=0; i<nbLigneMatrice; i++){
        for(int j=0; j<nbColonneMatrice; j++){
            matriceMines[i][j] = malloc(sizeof(struct celluleDemineur));
            matriceMines[i][j]->hasMine = false;
            matriceMines[i][j]->isPressed = false;
            matriceMines[i][j]->ifFlagged = false;
            matriceMines[i][j]->isNeareast = false;
            matriceMines[i][j]->nearbyMine = 0;
            matriceMines[i][j]->line = i;
            matriceMines[i][j]->column = j;
            matriceMines[i][j]->couleurCase = DARKGRAY;
            
            if(i%2 == 0){
                matriceMines[i][j]->pos = (Vector2){(1.5*j+1.75)*SIZE_OF_HEX + screenWidth/nbColonneMatrice, (1.5*i+1)*SIZE_OF_HEX + screenHeight/nbLigneMatrice};
            }else{
                matriceMines[i][j]->pos = (Vector2){(1.5*j+1)*SIZE_OF_HEX + screenWidth/nbColonneMatrice, (1.5*i+1)*SIZE_OF_HEX + screenHeight/nbLigneMatrice};
            }
        }
    }
}

float getDistanceBetweenPoints(Vector2 a, Vector2 b) {
  return (a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y);
}

int main(void)
{
    // Initialization
    //--------------------------------------------------------------------------------------
    const int screenWidth = 1280;
    const int screenHeight = 800;
    InitWindow(screenWidth, screenHeight, "HexaSweeper - Erwan MERLY");
    SetTargetFPS(60);
    
    srand(time(NULL));
    
    int nbLigneMatrice = 17, nbColonneMatrice = 21, nbMines = 80;
    int curCaseX = 0, curCaseY = 0, flaggedMines = 0;
    bool lose = false, win = false;
    float timeEllapsed = 0;
    
    //CREATION DE LA MATRICE DES MINES
    CelluleDemineur** matriceMines = malloc(nbLigneMatrice * sizeof(CelluleDemineur *));
    for(int i=0; i<nbLigneMatrice; i++)
        matriceMines[i] = calloc(nbLigneMatrice * nbColonneMatrice, sizeof(CelluleDemineur));
    initGrille(nbLigneMatrice, nbColonneMatrice, matriceMines);
    generationMines(nbMines, nbLigneMatrice, nbColonneMatrice, matriceMines);
    //--------------------------------------------------------------------------------------

    // Main game loop
    while (!WindowShouldClose())    // Detect window close button or ESC key
    {
        // Update
        //----------------------------------------------------------------------------------
        if(!win && !lose){
            timeEllapsed += GetFrameTime();
            
            if(IsMouseButtonPressed(MOUSE_LEFT_BUTTON)){
                lose = revelerCase(matriceMines, curCaseX, curCaseY, nbLigneMatrice, nbColonneMatrice);
            }
            
            if(IsMouseButtonPressed(MOUSE_RIGHT_BUTTON)){
                if(!matriceMines[curCaseX][curCaseY]->isPressed){
                    matriceMines[curCaseX][curCaseY]->ifFlagged = !matriceMines[curCaseX][curCaseY]->ifFlagged;
                    if(matriceMines[curCaseX][curCaseY]->ifFlagged){
                        matriceMines[curCaseX][curCaseY]->couleurCase = GREEN;
                        flaggedMines++;
                    }
                    else{
                        matriceMines[curCaseX][curCaseY]->couleurCase = DARKGRAY;
                        flaggedMines--;
                    }
                }
            }
            
            if(IsMouseButtonPressed(MOUSE_MIDDLE_BUTTON)){
                lose = revelerCase(matriceMines, curCaseX, curCaseY, nbLigneMatrice, nbColonneMatrice);
                lose = revelerVoisins(curCaseX, curCaseY, matriceMines, nbLigneMatrice, nbColonneMatrice);
            }
        }else{
            if(IsKeyPressed('R')){
                lose = false;
                win = false;
                flaggedMines = 0;
                timeEllapsed = 0;
                initGrille(nbLigneMatrice, nbColonneMatrice, matriceMines);
                generationMines(nbMines, nbLigneMatrice, nbColonneMatrice, matriceMines);
            }
        }
        
        Vector2 mPos = GetMousePosition();
        for(int i=0; i<nbLigneMatrice; i++){
            for(int j=0; j<nbColonneMatrice; j++){
                float distanceCur = getDistanceBetweenPoints(mPos, matriceMines[i][j]->pos);
                if(distanceCur < SIZE_OF_HEX*SIZE_OF_HEX/2){
                    matriceMines[i][j]->isNeareast = true;
                    curCaseX = matriceMines[i][j]->line;
                    curCaseY = matriceMines[i][j]->column;
                }else{
                    matriceMines[i][j]->isNeareast = false;
                }
            }
        }
        
        if(flaggedMines == nbMines){
            int trueFlagged = 0;
            for(int i=0; i<nbLigneMatrice; i++){
                for(int j=0; j<nbColonneMatrice; j++){
                    if(matriceMines[i][j]->hasMine && matriceMines[i][j]->ifFlagged) trueFlagged++;
                }
            }
            
            if(trueFlagged == nbMines) win = true;
        }
        
        // Draw
        //----------------------------------------------------------------------------------
        BeginDrawing();

            ClearBackground(RAYWHITE);
            
            //AFFICHAGE DE LA GRILLE
            for(int i=0; i<nbLigneMatrice; i++){
                for(int j=0; j<nbColonneMatrice; j++){
                    DrawPoly((Vector2){matriceMines[i][j]->pos.x, matriceMines[i][j]->pos.y}, 6, SIZE_OF_HEX, 0, GRAY);
                    DrawPoly((Vector2){matriceMines[i][j]->pos.x, matriceMines[i][j]->pos.y}, 6, SIZE_OF_HEX-3, 0, Fade(matriceMines[i][j]->couleurCase, matriceMines[i][j]->isNeareast ? 0.5f : 1.0f));
                    
                    if(matriceMines[i][j]->isPressed && matriceMines[i][j]->nearbyMine>0) 
                        DrawText(TextFormat("%i", matriceMines[i][j]->nearbyMine), matriceMines[i][j]->pos.x - SIZE_OF_HEX/3, matriceMines[i][j]->pos.y - SIZE_OF_HEX/3, SIZE_OF_HEX*0.8, LIME);
                }
            }
            
            //AFFICHAGE DE L'INTERFACE
            DrawRectangle(0, 725, 1280, 75, Fade(LIGHTGRAY, 0.5f));
            DrawText(TextFormat("MINE FLAGGED: %d/%d", flaggedMines, nbMines), 50, 750, 35, GRAY);
            DrawText(TextFormat("Time: %d s", (int)timeEllapsed), 1030, 750, 35, GRAY);
            
            //AFFICHAGE CONTROLES ET REGLES
            DrawText("CONTROLS", 1005, 50, 35, DARKGRAY);
            DrawText("LEFT CLICK - REVEAL TILE", 930, 100, 18, LIGHTGRAY);
            DrawText("RIGHT CLICK - FLAG TILE", 930, 130, 18, LIGHTGRAY);
            DrawText("MIDDLE CLICK - REVEAL NEIGHBOURS", 930, 160, 18, LIGHTGRAY);
            
            DrawText("RULES", 1005, 210, 35, DARKGRAY);
            DrawText("NUMBER INDICATE MINES AROUND", 930, 260, 18, LIGHTGRAY);
            DrawText("FLAG ALL THE MINES TO WIN", 930, 290, 18, LIGHTGRAY);
            
            DrawText("TILES", 1005, 340, 35, DARKGRAY);
            DrawText("DARK GRAY - NON REVEALED", 930, 390, 18, LIGHTGRAY);
            DrawText("LIGHT GRAY - REVEALED", 930, 420, 18, LIGHTGRAY);
            DrawText("GREEN - FLAGGED", 930, 450, 18, LIGHTGRAY);
            DrawText("RED - MINE", 930, 480, 18, LIGHTGRAY);
            
            //AFFICHAGE FIN DE PARTIE
            if(lose || win){
                DrawRectangle(0, 440, 1280, 100, Fade(GRAY, 0.8f));
                if(lose){
                    DrawText("GAME OVER", 500, 465, 35, MAROON);
                    for(int i=0; i<nbLigneMatrice; i++){
                        for(int j=0; j<nbColonneMatrice; j++){
                            if(containsMine(i, j, matriceMines)) revelerCase(matriceMines, i, j, nbLigneMatrice, nbColonneMatrice);
                        }
                    }
                } 
                if(win) DrawText("YOU WIN!", 500, 465, 35, LIME);
                DrawText("PRESS R TO REPLAY", 500, 495, 30, DARKGRAY);
            }            
            
        EndDrawing();
        //----------------------------------------------------------------------------------
    }

    // De-Initialization
    //--------------------------------------------------------------------------------------
    CloseWindow();        // Close window and OpenGL context
    
    for(int i=0; i<nbLigneMatrice; i++)
        free(matriceMines[i]);
    free(matriceMines);
    //--------------------------------------------------------------------------------------

    return 0;
}